import asyncio
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple, Union
from dokker.types import LogStream


class CommandError(Exception):
    pass


async def _aread_stream(
    stream: asyncio.StreamReader,
    queue: asyncio.Queue,
    name: str,
) -> None:
    async for line in stream:
        await queue.put((name, line.decode("utf-8").strip()))

    await queue.put(None)


async def astream_command(command: List[str]) -> LogStream:
    # Create the subprocess using asyncio's subprocess

    full_cmd = " ".join(map(str, command))

    proc = await asyncio.create_subprocess_shell(
        full_cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    if proc.stdout is None or proc.stderr is None:
        raise CommandError("Could not create the subprocess.")

    queue = asyncio.Queue()  # type: asyncio.Queue[tuple[str, str]]

    # cannot use type annotation because of python 3.8
    # Create and start tasks for reading each stream

    try:
        collected_logs = []
        readers = [
            asyncio.create_task(_aread_stream(proc.stdout, queue, "STDOUT")),
            asyncio.create_task(_aread_stream(proc.stderr, queue, "STDERR")),
        ]

        # Track the number of readers that are finished
        finished_readers = 0
        while finished_readers < len(readers):
            line = await queue.get()
            if line is None:
                finished_readers += 1  # One reader has finished
                continue
            collected_logs.append(line[1])
            yield line

        # Cleanup: cancel any remaining reader tasks
        for reader in readers:
            reader.cancel()
            try:
                await reader
            except asyncio.CancelledError:
                pass

        await proc.wait()

        if proc.returncode != 0:
            logs = "\n".join(collected_logs) if collected_logs else "No Logs"

            raise CommandError(
                f"Command {full_cmd} failed with return code {proc.returncode}: \n{logs}"
            )

    except asyncio.CancelledError:
        # Handle cancellation request
        proc.kill()
        await proc.wait()  # Wait for the subprocess to exit after receiving SIGINT

        # Cleanup: cancel any remaining reader tasks
        for reader in readers:
            reader.cancel()
            try:
                await reader
            except asyncio.CancelledError:
                pass

        raise

    except Exception as e:
        raise e



from pydantic import BaseModel, Field
import logging


class LoggingContext:
    def __init__(self, logger: logging.Logger, status: str):
        self.logger = logger
        self.status = status

    def __aenter__(self):
        self.logger.log("INFO", f"Start... {self.status}")
        return self

    def __aexit__(self, exc_type, exc_val, exc_tb):
        self.logger.log("INFO", f"Done... {self.status}")
        pass


class Logger(BaseModel):
    """A logger that prints all logs to a logger"""

    logger: logging.Logger = Field(default_factory=lambda: logging.getLogger(__name__))
    log_level: int = logging.INFO

    def status(self, log: str) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """
        return LoggingContext(self.logger, log)

    def log(self, log: str) -> None:
        """A method for logs

        Parameters
        ----------
        log : str
            The log to print
        """

        self.logger.log(self.log_level, log)

    class Config:
        """pydantic config class"""

        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from typing import Protocol, runtime_checkable, Dict, Any
from .cli import CLI


@runtime_checkable
class Project(Protocol):
    """A Projectaget_values, it can decide to setup the project asy
    nchronousl, e.g by cloning a git repository, or copiying a
    directory into the .dokker directory. The project can also
    implement methods that will be run before and after certain
    docker-compose commands are run. For example, a project can
    implement a method that will be run before the docker-compose
    up command is run.
    """

    async def ainititialize(self) -> CLI:
        """A setup method for the project.

        Returns
        -------
        CLI
            The CLI to use for the project.
        """
        ...



    async def atear_down(self, cli: CLI) -> None:
        """Tear down the project.

        A project can implement this method to tear down the project
        when the project is torn down. This can be used to remove
        temporary files, or to remove the project from the .dokker
        directory.

        Parameters
        ----------
        cli : CLI
            The CLI that was used to run the project.

        """
        ...



    async def abefore_pull(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_up(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_enter(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_down(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_stop(self) -> None:
        """A selisttup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...



from pydantic import BaseModel, Field
import os
from typing import Optional
import shutil
from dokker.cli import CLI
from dokker.types import ValidPath


class CopyPathProject(BaseModel):
    """A copy path Project.

    This project is a project that will mirror a path to a temporary
    directory and run it from there. This is useful for testing projects that
    are in production environments btu should be tested locally.
    """
    project_path: ValidPath
    project_name: Optional[str] = None
    base_dir: str = Field(default_factory=lambda: os.path.join(os.getcwd(), ".dokker"))
    overwrite: bool = False

    async def ainititialize(self) -> CLI:
        """A setup method for the project.

        Returns
        -------
        CLI
            The CLI to use for the project.
        """
        os.makedirs(self.base_dir, exist_ok=True)

        if self.project_name is None:
            self.project_name = os.path.basename(self.project_path)

        project_dir = os.path.join(self.base_dir, self.project_name)
        if os.path.exists(project_dir) and not self.overwrite:
            raise Exception(
                f"Project {self.project_name} already exists in {self.base_dir}. Set overwrite to overwrite."
            )

        shutil.copytree(self.project_path, project_dir, dirs_exist_ok=self.overwrite)

        compose_file = os.path.join(project_dir, "docker-compose.yml")
        if not os.path.exists(compose_file):
            raise Exception(
                "No docker-compose.yml found in the template. It appears that the template is not a valid dokker template."
            )

        return CLI(compose_files=[compose_file])
    
    async def atear_down(self, cli: CLI) -> None:
        """Tear down the project.

        A project can implement this method to tear down the project
        when the project is torn down. This can be used to remove
        temporary files, or to remove the project from the .dokker
        directory.

        Parameters
        ----------
        cli : CLI
            The CLI that was used to run the project.

        """
        


        if self.project_name is None:
            self.project_name = os.path.basename(self.project_path)


        project_dir = os.path.join(self.base_dir, self.project_name)
        if os.path.exists(project_dir):
            shutil.rmtree(project_dir)

    async def abefore_pull(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_up(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_enter(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_down(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_stop(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...



from .deployment import Deployment, HealthCheck
from typing import List, Optional, TYPE_CHECKING, Union
from dokker.projects.copy import CopyPathProject
from dokker.projects.local import LocalProject
from dokker.types import ValidPath

if TYPE_CHECKING:
    pass


def mirror(
      local_path: ValidPath, health_checks: Optional[List[HealthCheck]] = None  
) -> Deployment:
    """Creates a Mirro Deployment

    A mirror deployment is a deployment that copies a local path to a temporary
    directory and runs it from there. This is useful for testing projects that
    are in production environments btu should be tested locally.

    Parameters
    ----------
    local_path : ValidPath
        _description_
    health_checks : Optional[List[HealthCheck]], optional
        _description_, by default None

    Returns
    -------
    Deployment
        The deployment
    """
    if health_checks is None:
        health_checks = []
    
    project =  CopyPathProject(project_path=local_path)
    deployment = Deployment(
        project=project,
        health_checks=health_checks,
    )

    deployment.pull_on_enter = False
    deployment.down_on_exit = False
    deployment.stop_on_exit = False

    return deployment



def cookiecutter(
        repo_url: str, health_checks: Optional[List[HealthCheck]] = None
) -> Deployment:
    """Generates a cookiecutter deployemnt.

    A cookiecutter deployemnt is a deployemnt that is generated from a cookiecutter template.
    This function is a helper function to generate a CookieCutterProject, and
    requires that cookiecutter is installed.

    Parameters
    ----------
    repo_url : str
        The url to the cookiecutter template.

    Returns
    -------
    Deployment
        The deployment
    """
    from dokker.projects.contrib.cookiecutter import CookieCutterProject

    if health_checks is None:
        health_checks = []

    project =  CookieCutterProject(repo_url=repo_url)
    deployment = Deployment(
        project=project,
        health_checks=health_checks,
    )

    deployment.pull_on_enter = False
    deployment.down_on_exit = False
    deployment.stop_on_exit = False

    return deployment


def local(
    docker_compose_file: Union[ValidPath, List[ValidPath]], health_checks: Optional[List[HealthCheck]] = None
) -> Deployment:
    """Creates a local deployment.

    A local deployment is a deployment that runs a docker-compose file
    locally. This is useful for testing a deployment that we do not want to
    control (e.g. calling down) on exit.
    """
    if not isinstance(docker_compose_file, list):
        docker_compose_file = [docker_compose_file]

    if health_checks is None:
        health_checks = []
        
    project = LocalProject(
        compose_files=docker_compose_file,
    )
    deployment = Deployment(
        project=project,
        health_checks=health_checks,
    )

    deployment.pull_on_enter = False
    deployment.down_on_exit = False
    deployment.stop_on_exit = True
    deployment.initialize_on_enter = True
    deployment.up_on_enter = True
    deployment.down_on_exit = False
    return deployment


def monitoring(
    docker_compose_file: Union[ValidPath, List[ValidPath]], health_checks: Optional[List[HealthCheck]] = None
) -> Deployment:
    """Generates a monitoring deployment.

    A monitoring deployment is a deployment that never directly interacts with the
    docker-compose CLI. This is useful for inspect / monitoring a deployment
    that is running in production.

    Parameters
    ----------
    docker_compose_file : Union[ValidPath, List[ValidPath]]
        The docker-compose file to run.
    health_checks : Optional[List[HealthCheck]], optional
        The health checks to run, by default None

    Returns
    -------
    Deployment
        The deployment
    """
    if not isinstance(docker_compose_file, list):
        docker_compose_file = [docker_compose_file]
    if health_checks is None:
        health_checks = []
    project = LocalProject(
        compose_files=docker_compose_file,
    )
    deployment = Deployment(
        project=project,
        health_checks=health_checks,
    )

    deployment.pull_on_enter = False
    deployment.down_on_exit = False
    deployment.stop_on_exit = False
    deployment.health_on_enter = True
    deployment.up_on_enter = False

    return deployment


def testing(
    docker_compose_file: Union[ValidPath, List[ValidPath]], health_checks: Optional[List[HealthCheck]] = None
) -> Deployment:
    """Generates a testing deployment.

    A testing deployment is a deployment that runs a docker-compose file, locally
    and takes care of pulling, initializing, and tearing down the deployment.

    Parameters
    ----------
    docker_compose_file : Union[ValidPath, List[ValidPath]]
        The docker-compose file to run.
    health_checks : Optional[List[HealthCheck]], optional
        The health checks to run, by default None

    Returns
    -------
    Deployment
        The deployment
    """
    if not isinstance(docker_compose_file, list):
        docker_compose_file = [docker_compose_file]
    if health_checks is None:
        health_checks = []
    project = LocalProject(
        compose_files=docker_compose_file,
    )
    deployment = Deployment(
        project=project,
        health_checks=health_checks,
    )

    deployment.pull_on_enter = True
    deployment.initialize_on_enter = True
    deployment.up_on_enter = True
    deployment.down_on_exit = True
    deployment.stop_on_exit = True
    deployment.tear_down_on_exit = True

    return deployment



class DokkerError(Exception):
    """Base class for all Dokker errors."""


class NotInitializedError(DokkerError):
    """Raised when Dokker is not initialized."""


class NotInspectedError(DokkerError):
    """Raised when an object is not inspected."""


class NotInspectableError(DokkerError):
    """Raised when an object is not inspectable."""




from cookiecutter.main import cookiecutter
from pydantic import BaseModel, Field
import os
from dokker.cli import CLI
from typing import Optional, Dict, Any
from dokker.errors import DokkerError
import shutil
import yaml

class CookieNotInstalled(DokkerError):
    """Raised when cookiecutter was instructed to tear down a project, but the project was not initialized."""
    pass


class CookieCutterProject(BaseModel):
    """A project that is generated from a cookiecutter template.

    This project is a project that is generated from a cookiecutter template.
    It can be used to run a docker-compose file locally, copying the template
    to the .dokker directory, and running the docker-compose file from there.
    """

    repo_url: str
    base_dir: str = Field(default_factory=lambda: os.path.join(os.getcwd(), ".dokker"))
    compose_files: list = Field(default_factory=lambda: ["docker-compose.yml"])
    extra_context: dict = Field(default_factory=lambda: {})
    overwrite_if_exists: bool = False

    _project_dir: Optional[str] = None
    _outs: Optional[Dict[str, Any]] = None

    async def ainititialize(self) -> CLI:
        """A setup method for the project.

        Returns
        -------
        CLI
            The CLI to use for the project.
        """
        os.makedirs(self.base_dir, exist_ok=True)

        self._project_dir = cookiecutter(
            self.repo_url,
            no_input=True,
            output_dir=self.base_dir,
            extra_context=self.extra_context,
            overwrite_if_exists=self.overwrite_if_exists,
        )
        assert isinstance(self._project_dir, str), "cookiecutter should return a string"

        compose_file = os.path.join(self._project_dir, "docker-compose.yml")
        if not os.path.exists(compose_file):
            raise Exception(
                "No docker-compose.yml found in the template. It appears that the template is not a valid dokker template."
            )

        return CLI(compose_files=[compose_file])
    
        
        
        
        
    

    
    async def atear_down(self, cli: CLI) -> None:
        """Tear down the project.

        A project can implement this method to tear down the project
        when the project is torn down. This can be used to remove
        temporary files, or to remove the project from the .dokker
        directory.

        Parameters
        ----------
        cli : CLI
            The CLI that was used to run the project.

        """
        if not self._project_dir:
            raise CookieNotInstalled("Cookiecutter project not installed. Did you call initialize?")
        
        if os.path.exists(self._project_dir):
            shutil.rmtree(self._project_dir)

    async def abefore_pull(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_up(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_enter(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_down(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_stop(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    class Config:
        """pydantic config class for CookieCutterProject"""
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True


from pydantic import BaseModel, Field
from typing import Dict, Optional, List, Protocol, runtime_checkable
from koil.composition import KoiledModel
import asyncio
from pathlib import Path
from dokker.compose_spec import ComposeSpec
from dokker.project import Project
from typing import Union
from koil import unkoil
from dokker.cli import CLI
from dokker.loggers.void import VoidLogger
from .log_watcher import LogWatcher
import aiohttp
import certifi
from ssl import SSLContext
import ssl
from typing import Any, Optional, List, Union
from dokker.errors import NotInitializedError, NotInspectedError


ValidPath = Union[str, Path]


class HealthError(Exception):
    pass


class HealthCheck(BaseModel):
    url: str
    service: str
    max_retries: int = 3
    timeout: int = 10
    error_with_logs: bool = True
    headers: Optional[dict] = Field(
        default_factory=lambda: {"Content-Type": "application/json"}
    )
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        description="SSL Context to use for the request",
    )

    async def acheck(self):
        async with aiohttp.ClientSession(
            headers=self.headers,
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:
            # get json from endpoint
            async with session.get(self.url) as resp:
                assert resp.status == 200
                return await resp.text()

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True


@runtime_checkable
class Logger(Protocol):
    def on_pull(self, log: str):
        ...

    def on_up(self, log: str):
        ...

    def on_stop(self, log: str):
        ...

    def on_logs(self, log: str):
        ...

    def on_down(self, log: str):
        ...


class Deployment(KoiledModel):
    """A deployment is a set of services that are deployed together."""

    project: Project = Field(default_factory=Project)
    services: Optional[List[str]] = None

    health_checks: List[HealthCheck] = Field(default_factory=list)
    initialize_on_enter: bool = False
    inspect_on_enter: bool = False
    pull_on_enter: bool = False
    up_on_enter: bool = False
    health_on_enter: bool = False
    down_on_exit: bool = False
    stop_on_exit: bool = False
    tear_down_on_exit: bool = False
    threadpool_workers: int = 3

    pull_logs: Optional[List[str]] = None
    up_logs: Optional[List[str]] = None
    stop_logs: Optional[List[str]] = None

    auto_initialize: bool = True

    logger: Logger = Field(default_factory=VoidLogger)

    _spec: ComposeSpec = None
    _cli: CLI = None

    @property
    def spec(self) -> ComposeSpec:
        """A property that returns the compose spec of the deployment.

        THis compose spec can be used to retrieve information about the deployment.
        by inspecting the containers, networks, volumes, etc.

        In the future, this spec will be used to
        retrieve information about the deployment.

        Returns
        -------
        ComposeSpec
            The compose spec.

        Raises
        ------
        NotInspectedError
            If the deployment has not been inspected.
        """
        if self._spec is None:
            raise NotInspectedError(
                "Deployment not inspected. Call await deployment.ainspect() first."
            )
        return self._spec

    async def ainitialize(self) -> "CLI":
        """Initialize the deployment.

        Will initialize the deployment through its project and return the CLI object.
        This method is called automatically when using the deployment as a context manager.

        Returns
        -------
        CLI
           The CLI object.
        """
        self._cli = await self.project.ainititialize()
        return self._cli

    async def aretrieve_cli(self):
        if self._cli is None:
            if self.auto_initialize:
                await self.ainitialize()
            else:
                raise NotInitializedError(
                    "Deployment not initialized and auto_initialize is False. Call await deployment.ainitialize() first."
                )

        return self._cli

    async def ainspect(self) -> ComposeSpec:
        """Inspect the deployment.

        Will inspect the deployment through its project and return the compose spec, which
        can be used to retrieve information about the deployment.
        This method is called automatically when using the deployment as a context manager and
        if inspect_on_enter is True.
        Returns
        -------
        ComposeSpec
            The compose spec.

        Raises
        ------
        NotInitializedError
            If the deployment has not been initialized.
        """
        cli = await self.aretrieve_cli()
        self._spec = await cli.ainspect_config()
        return self._spec

    def inspect(self) -> ComposeSpec:
        return unkoil(self.ainspect)

    def add_health_check(
        self,
        url: str,
        service: str,
        max_retries: int = 3,
        timeout: int = 10,
        error_with_logs: bool = True,
    ) -> "HealthCheck":
        """Add a health check to the deployment.

        Parameters
        ----------
        url : str
            The url to check.
        service : str
            The service this health check is for.
        max_retries : int, optional
            The maximum retries before the healtch checks fails, by default 3
        timeout : int, optional
            The timeout between retries, by default 10
        error_with_logs : bool, optional
            Should we error with the logs of the service (will inspect container logs of the service), by default True

        Returns
        -------
        HealthCheck
            The health check object.
        """

        check = HealthCheck(
            url=url,
            service=service,
            max_retries=max_retries,
            timeout=timeout,
            error_with_logs=error_with_logs,
        )

        self.health_checks.append(check)
        return check

    async def acheck_healthz(self, check: HealthCheck, retry: int = 0):
        try:
            await check.acheck()
        except Exception as e:
            if retry < check.max_retries:
                await asyncio.sleep(check.timeout)
                await self.acheck_healthz(check, retry=retry + 1)
            else:
                if not check.error_with_logs:
                    raise HealthError(
                        f"Health check failed after {check.max_retries} retries. Logs are disabled."
                    ) from e

                logs = []

                async for std, i in self._cli.astream_docker_logs(
                    services=[check.service]
                ):
                    logs.append(i)

                raise HealthError(
                    f"Health check failed after {check.max_retries} retries. Logs:\n"
                    + "\n".join(logs)
                ) from e

    async def await_for_healthz(
        self, timeout: int = 3, retry: int = 0, services: Optional[List[str]] = None
    ):
        if services is None:
            services = [
                check.service for check in self.health_checks
            ]  # we check all services

        return await asyncio.gather(
            *[
                self.acheck_healthz(check)
                for check in self.health_checks
                if check.service in services
            ]
        )

    def logswatcher(self, service_names: Union[List[str], str], **kwargs):
        """Get a logswatcher for a service.

        A logswatcher is an object that can be used to watch the logs of a service, as
        they are being streamed. It is an (async) context manager that should be used
        to enclose any code that needs to watch the logs of a service.

        ```python
         with deployment.logswatcher("service"):
            # do something with service logs
            print(requests.get("http://service").text

        ```

        If you want to watch the logs of multiple services, you can pass a list of service names.

            ```python

            watcher = deployment.logswatcher(["service1", "service2"])
            with watcher:
                # do something with service logs
                print(requests.get("http://service1").text
                print(requests.get("http://service2").text

            print(watcher.collected_logs)

            ```

        Parameters
        ----------
        service_name : Union[List[str], str]
            The name of the service(s) to watch the logs for.

        Returns
        -------
        LogWatcher
            The log watcher object.
        """
        if isinstance(service_names, str):
            service_names = [service_names]

        return LogWatcher(cli_bearer=self, services=service_names, tail=1, **kwargs)

    async def aup(self, detach=True):
        """Up the deployment.

        Will call docker-compose up on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if up_on_enter is True.

        Returns
        -------
        List[str]
            The logs of the up command.
        """
        cli = await self.aretrieve_cli()
        logs = []
        async for type, log in cli.astream_up(detach=detach):
            logs.append(log)
            self.logger.on_up(log)

        return logs

    def up(self, detach=True):
        """Up the deployment.

        Will call docker-compose up on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if up_on_enter is True.

        Returns
        -------
        List[str]
            The logs of the up command.
        """

        return unkoil(self.aup, detach=detach)

    async def arestart(
        self,
        services: Union[List[str], str],
        await_health: bool = True,
        await_health_timeout: int = 3,
    ):
        """Restarts a service.

        Will call docker-compose restart on the list of services.
        If await_health is True, will await for the health checks of these services to pass.

        Parameters
        ----------
        services : Union[List[str], str], optional
            The list of services to restart, by default None
        await_health : bool, optional
            Should we await for the health checks to pass, by default True
        await_health_timeout : int, optional
            The time to wait for  before checking the health checks (allows the container to
            shutdown), by default 3, is void if await_health is False

        Returns
        -------
        List[str]
            The logs of the restart command.
        """
        cli = await self.aretrieve_cli()
        if isinstance(services, str):
            services = [services]

        logs = []
        async for type, log in cli.astream_restart(services=services):
            logs.append(log)

        if await_health:
            await asyncio.sleep(await_health_timeout)
            await self.await_for_healthz(services=services)

        return logs

    def restart(
        self,
        services: Union[List[str], str],
        await_health: bool = True,
        await_health_timeout: int = 3,
    ):
        """Restarts a service. (sync)

        Will call docker-compose restart on the list of services.
        If await_health is True, will await for the health checks of these services to pass.

        Parameters
        ----------
        services : Union[List[str], str], optional
            The list of services to restart, by default None
        await_health : bool, optional
            Should we await for the health checks to pass, by default True
        await_health_timeout : int, optional
            The time to wait for  before checking the health checks (allows the container to
            shutdown), by default 3, is void if await_health is False

        Returns
        -------
        List[str]
            The logs of the restart command.
        """
        return unkoil(
            self.arestart,
            services=services,
            await_health=await_health,
            await_health_timeout=await_health_timeout,
        )

    async def apull(self):
        """Pull the deployment.

        Will call docker-compose pull on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if pull_on_enter is True.

        Returns
        -------
        List[str]
            The logs of the pull command.

        Raises
        ------
        NotInitializedError
            If the deployment has not been initialized.
        """
        cli = await self.aretrieve_cli()

        logs = []
        async for type, log in cli.astream_pull():
            logs.append(log)
            self.logger.on_pull(log)

        return logs

    async def adown(self) -> List[str]:
        """Down the deployment.

        Will call docker-compose down on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if down_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the down command.
        """
        cli = await self.aretrieve_cli()

        logs = []
        async for type, log in cli.astream_down():
            logs.append(log)
            self.logger.on_down(log)

        return logs

    async def aremove(self) -> None:
        """Down the deployment.

        Will call docker-compose down on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if down_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the down command.
        """
        cli = await self.aretrieve_cli()

        return await self.project.atear_down(cli)

    def remove(self) -> None:
        """Remove the project

        Returns
        -------
        List[str]
            The logs of the down command.
        """
        return unkoil(self.aremove)

    def down(self) -> List[str]:
        """Down the deployment.

        Will call docker-compose down on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if down_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the down command.
        """
        return unkoil(self.adown)

    async def astop(self) -> List[str]:
        """Stop the deployment.

        Will call docker-compose stop on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if stop_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the stop command.
        """
        cli = await self.aretrieve_cli()

        logs = []
        async for type, log in cli.astream_stop():
            logs.append(log)
            self.logger.on_stop(log)

        return logs

    def stop(self) -> List[str]:
        """Stop the deployment.

        Will call docker-compose stop on the deployment.
        This method is called automatically when using the deployment as a context manager and
        if stop_on_exit is True.

        Returns
        -------
        List[str]
            The logs of the stop command.
        """
        return unkoil(self.astop)

    async def aget_cli(self):
        """Get the CLI object of the deployment.

        THis is the defining method of a CLI bearer, and will
        be called by any method that needs the CLI object.
        This is an async method because initializing the CLI object
        is an async operation (as it might incure network calls).
        """
        return await self.aretrieve_cli()

    async def __aenter__(self) -> "Deployment":
        """Async enter method for the deployment.

        Will initialize the project, if auto_initialize is True.
        Will inspect the deployment, if inspect_on_enter is True.
        Will call docker-compose up and pull on the deployment, if
        up_on_enter and pull_on_enter are True respectively.

        """
        if self.initialize_on_enter:
            await self.ainitialize()

        if self.inspect_on_enter:
            await self.ainspect()

        if self.pull_on_enter:
            await self.project.abefore_pull()
            await self.apull()

        if self.up_on_enter:
            await self.project.abefore_up()
            await self.aup()

        if self.health_on_enter:
            if self.health_checks:
                await self.await_for_healthz()

        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async exit method for the deployment.

        Will call docker-compose down and stop on the deployment, if
        down_on_exit and stop_on_exit are True respectively.
        """
        if self.stop_on_exit:
            await self.project.abefore_stop()
            await self.astop()

        if self.down_on_exit:
            await self.project.abefore_down()
            await self.adown()

        if self.tear_down_on_exit:
            await self.project.atear_down(self._cli)

        self._cli = None

    class Config:
        """Pydantic configuration for the deployment."""

        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



from pydantic import BaseModel, Field
import os
from dokker.cli import CLI, LogStream
from typing import Optional, List
from dokker.errors import DokkerError
import shutil
import asyncio
import ssl
import certifi
from ssl import SSLContext
import aiohttp
import json
import yaml
from typing import Dict, Any, Protocol, runtime_checkable
from aioconsole import ainput
from enum import Enum
from pydantic import validator
from dokker.command import astream_command

class InitError(DokkerError):
    """Raised when cookiecutter was instructed to tear down a project, but the project was not initialized."""

    pass


class BasicType(str, Enum):
    STR = "str"
    INT = "in self.t"
    FLOAT = "float"
    BOOL = "bool"
    LIST = "list"



def set_nested_key_in_dict(d, key, value):
    if "." not in key:
        d[key] = value
        return

    key, rest = key.split(".", 1)
    if key not in d:
        d[key] = {}
    set_nested_key_in_dict(d[key], rest, value)

def get_nested_key_in_dict(d, key):
    if "." not in key:
        return d.get(key, None)

    key, rest = key.split(".", 1)
    return get_nested_key_in_dict(d.get(key, {}), rest)



class BasicField(BaseModel):
    key: str
    label: str
    description: Optional[str] = ""
    type: BasicType = BasicType.STR
    required: bool = False
    child: Optional["BasicField"] = None
    choices: Optional[List[Any]] = None

    async def aprompt(self: "BasicField", default: Dict[str, Any]) -> Dict[str, Any]:
        if self.type == BasicType.LIST:
            if self.child is None:
                raise Exception("Child field is required for list type")

            default_values = default.get(self.key, [])
            if not isinstance(default_values, list):
                raise Exception("Default value must be a list")

            if len(default_values) > 0:
                print(f"Current values: {default_values}")
                if await ainput("Change values? [y/n]: ") != "y":
                    return default

            list_values = []

            while True:
                # Default values do not work for child fields
                list_values.append(await self.child.aprompt({}))
                if await ainput("Add another? [y/n]: ") != "y":
                    break

            new_defaults = {**default}
            set_nested_key_in_dict(new_defaults, self.key, list_values)
            return new_defaults

        default_value: str = get_nested_key_in_dict(default, self.key)

        prompt_string = f"{self.label}: {self.description} "
        if default_value is not None:
            prompt_string += f" [{default_value}] "

        value = await ainput(prompt_string)
        if value == "" and default_value is not None:
            value = default_value
        else:

            if self.type == BasicType.INT:
                value = int(value)
            elif self.type == BasicType.FLOAT:
                value = float(value)
            elif self.type == BasicType.BOOL:
                value = value.lower() == "true"
            elif self.type == BasicType.LIST:
                value = value.split(",")
            elif self.type == BasicType.STR:
                value = str(value)
            else:
                raise Exception("Unknown type")
        



        new_defaults = {**default}
        set_nested_key_in_dict(new_defaults, self.key, value)
        return new_defaults

    
    


class BasicForm(BaseModel):
    welcome_message: str = "Please fill out the following fields:"
    fields: List[BasicField] = Field(default_factory=lambda: [])

    async def aretrieve(self, default: Dict[str, Any]) -> Dict[str, Any]:
        print(self.welcome_message)

        for field in self.fields:
            default = await field.aprompt(default)

        return default


class Feature(BaseModel):
    name: str
    description: str
    default: bool = False


class Channel(BaseModel):
    name: str
    title: str
    experimental: bool = False
    logo: Optional[str] = None
    long: Optional[str] = None
    description: Optional[str] = None
    features: List[Feature] = []
    preview: bool = False
    builder: str
    forms: List[str] = []
    defaults: dict = {}
    wizard: List[BasicForm] = []
    basic_forms: List[BasicForm] = []


class RepoModel(BaseModel):
    repo: str
    channels: List["Channel"]


@runtime_checkable
class Form(Protocol):
    async def aretrieve(self, default: Dict[str, Any]) -> Dict[str, Any]:
        ...


class FormRegistry(BaseModel):
    registered_forms: Dict[str, Form] = {}

    def register(self, name: str, form: Form) -> None:
        self.registered_forms[name] = form

    def get(self, name: str) -> Optional[Form]:
        try:
            return self.registered_forms[name]
        except KeyError:
            return None

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True


@runtime_checkable
class Repo(Protocol):
    async def aload(self) -> RepoModel:
        ...


class RemoteRepo(BaseModel):
    url: str
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        description="SSL Context to use for the request",
    )
    headers: Optional[dict] = Field(
        default_factory=lambda: {"Content-Type": "application/json"}
    )

    async def aload(self) -> RepoModel:
        async with aiohttp.ClientSession(
            headers=self.headers,
            connector=aiohttp.TCPConnector(ssl=self.ssl_context),
        ) as session:
            # get json from endpoint
            async with session.get(self.url) as resp:
                assert resp.status == 200

                raw_json = await resp.text()
                json_data = json.loads(raw_json)
                return RepoModel(**json_data)

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True


class ModelRepo(BaseModel):
    model: RepoModel

    async def aload(self) -> RepoModel:
        return self.model


class KonstruktorProject(BaseModel):
    """A project that is generated from a cookiecutter template.

    This project is a project that is generated from a cookiecutter template.
    It can be used to run a docker-compose file locally, copying the template
    to the .dokker directory, and running the docker-compose file from there.
    """

    channel: str = "paper"
    repo: Repo
    base_dir: str = Field(default_factory=lambda: os.path.join(os.getcwd(), ".dokker"))
    compose_files: list = Field(default_factory=lambda: ["docker-compose.yml"])
    extra_context: dict = Field(default_factory=lambda: {})
    error_if_exists: bool = True
    reinit_if_exists: bool = False
    ssl_context: SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        description="SSL Context to use for the request",
    )
    headers: Optional[dict] = Field(
        default_factory=lambda: {"Content-Type": "application/json"}
    )
    name: Optional[str] = None
    skip_forms: bool = False
    form_registry: FormRegistry = Field(default_factory=lambda: FormRegistry())

    _project_dir: Optional[str] = None
    _outs: Optional[Dict[str, Any]] = None

    @validator("repo", pre=True)
    def validate_repo(cls, v) -> Repo:
        """Validate the repo type."""

        if isinstance(v, RepoModel):
            return ModelRepo(model=v)
        elif isinstance(v, str):
            return RemoteRepo(url=v)
        elif isinstance(v, Repo):
            return v
        else:
            raise ValueError("Invalid repo type")


    async def fetch_image(self, image: str) -> List[str]:
        logs: List[str] = []

        async for type, log in astream_command(["docker", "pull", image]):
            logs.append(log)

        return logs

    async def arun_form(self, form: str, defaults: Dict[str, Any]) -> Dict[str, Any]:
        form = self.form_registry.get(form)
        if form is None:
            return defaults
        return await form.aretrieve(defaults)

    async def ainititialize(self) -> CLI:
        """A setup method for the project.

        Returns
        -------
        CLI
            The CLI to use for the project.
        """

        repo = await self.repo.aload()

        try:
            channel = next(filter(lambda x: x.name == self.channel, repo.channels))
        except StopIteration:
            raise InitError(
                f"Channel {self.channel} not found in repo. Available are {', '.join(map(lambda x: x.name, repo.channels))}"
            )

        os.makedirs(self.base_dir, exist_ok=True)

        project_name = self.name or channel.name
        self._project_dir = os.path.join(self.base_dir, project_name)

        if os.path.exists(self._project_dir):
            if self.error_if_exists and not self.reinit_if_exists:
                raise InitError("Project already exists.")

            if not self.reinit_if_exists:
                print("Project already exists. Skipping initialization.")
                compose_file = os.path.join(self._project_dir, "docker-compose.yaml")
                if not os.path.exists(compose_file):
                    raise Exception(
                        "No docker-compose.yml found in the template. It appears that the template is not a valid dokker template. User overwrite_if_exists to overwrite the project."
                    )

                return CLI(
                    compose_files=[compose_file],
                )
            else:
                print("Project already exists. Overwriting.")
                shutil.rmtree(self._project_dir)

        setup_dict = {**channel.defaults, **self.extra_context}

        if not self.skip_forms:
            for form in channel.forms:
                setup_dict = await self.arun_form(form, setup_dict)

            for basic_form in channel.basic_forms:
                setup_dict = await basic_form.aretrieve(setup_dict)

        print("Fetching builder image...")
        logs = await self.fetch_image(channel.builder)

        os.makedirs(self._project_dir, exist_ok=True)
        # create setup.yaml
        setup_yaml = os.path.join(self._project_dir, "setup.yaml")

        with open(setup_yaml, "w") as f:
            yaml.dump(setup_dict, f)

        logs = []

        cmd = [
            "docker",
            "run",
            "--rm",
            "-v",
            f"{self._project_dir}:/app/init",
        ]

        if os.name == "posix":
            cmd += ["--user", f"{os.getuid()}:{os.getgid()}"]

        cmd += [channel.builder]

        async for type, log in astream_command(cmd):
            logs.append(log)

        compose_file = os.path.join(self._project_dir, "docker-compose.yaml")
        if not os.path.exists(compose_file):
            raise Exception(
                "No docker-compose.yml found in the template. It appears that the template is not a valid dokker template."
            )

        return CLI(
            compose_files=[compose_file],
        )

    async def atear_down(self, cli: CLI) -> None:
        """Tear down the project.

        A project can implement this method to tear down the project
        when the project is torn down. This can be used to remove
        temporary files, or to remove the project from the .dokker
        directory.

        Parameters
        ----------
        cli : CLI
            The CLI that was used to run the project.

        """
        try:
            await cli.adown()
        except Exception as e:
            print(e)
            pass

        if not self._project_dir:
            raise InitError(
                "Cookiecutter project not installed. Did you call initialize?"
            )

        print("Removing project directory...")
        if os.path.exists(self._project_dir):
            shutil.rmtree(self._project_dir)

        print("Removed project directory.")

    async def abefore_pull(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_up(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_enter(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_down(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    async def abefore_stop(self) -> None:
        """A setup method for the project.

        Returns:
            Optional[List[str]]: A list of logs from the setup process.
        """
        ...

    class Config:
        """pydantic config class for CookieCutterProject"""

        arbitrary_types_allowed = True
        underscore_attrs_are_private = True
        extra = "forbid"
